﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using PAT.Common.Classes.Expressions.ExpressionClass;
namespace PAT.Lib
{
    class Processor : ExpressionValue
    {
        public enum State { Inactive, Active };
        public enum Mode { Legacy, Guest, Host };
        public enum Reg { guestPTPointer, ProgramPointer, HypervisorEntry };
        public enum VMExitReason { Default, Hypercall, PrivilegedInstruction, Error };


        State _State = State.Inactive;      //Processor State
        Mode _Mode = Mode.Legacy;           //Processor Mode

        int _pagesize = 8;                   //MMU page size


        //guest states
        Prog _guestCachedProgram = null;     //program id
        int _guestNextPP = -1;            //virtual address
        int _guestPTP = -1;             //paging table used in guest mode, physical address
        int _guestPP = -1;              //virtual address
        int _guestHVEntry = -1;

        //host (or legacy) states
        Prog _hostCachedProgram = null;     //CSP shared index
        int _hostNextPP = -1;            //VA==PA
        int _hostPP = -1;               //VA=PA


        //sync states
        SyncException.SyncReason syncstate = SyncException.SyncReason.Nothing;
        Concat syncparam = null;
        //Concat syncsubparam = null;

        //vmexit
        VMExitReason vmexitreason = VMExitReason.Default;
        int hypercallid = 0;

        //Processor State
        public State state { get { return _State; } }
        public bool isActive { get { return state == State.Active; } }
        //Processor Mode
        public Mode mode { get { return _Mode; } }
        public bool inLegacy { get { return _Mode == Mode.Legacy; } }
        public bool inGuest { get { return _Mode == Mode.Guest; } }
        public bool inHost { get { return _Mode == Mode.Host; } }
        //memory page size
        public int pagesize { get { return _pagesize; } }

        //Guset/Host States
        public Prog cachedProgram
        {
            get
            {
                if (inGuest)
                    return _guestCachedProgram;
                else
                    return _hostCachedProgram;
            }
            set
            {
                if (inGuest)
                    _guestCachedProgram = value;
                else
                    _hostCachedProgram = value;
            }
        }

        public int nextPP
        {
            get
            {
                if (inGuest)
                    return _guestNextPP;
                else
                    return _hostNextPP;
            }
            set
            {
                if (inGuest)
                    _guestNextPP = value;
                else
                    _hostNextPP = value;
            }
        }

        public int PP
        {
            get
            {
                if (inGuest)
                    return _guestPP;
                else
                    return _hostPP;
            }
            set
            {
                if (inGuest)
                    _guestPP = value;
                else
                    _hostPP = value;
            }
        }

        //Guest States
        public int guestPTP { get { return _guestPTP; } }
        public int HVEntry { get { return _guestHVEntry; } }


        //Sync
        public int syncID { get { return (int)syncstate; } }
        public Concat syncParam { get { return syncparam; } }
        //public Concat syncSubParam { get { return syncsubparam; } }

        //VMExit
        public VMExitReason vmexitReason { get { return vmexitreason; } }
        public int hypercallID { get { return hypercallid; } }

        public Processor() { }

        public Processor(int __pagesize) { _pagesize = __pagesize; }


        public Processor(Processor obj)
        {
            _pagesize = obj._pagesize;
            _State = obj.state;
            syncstate = obj.syncstate;
            if (obj.syncparam != null) syncparam = (Concat)obj.syncparam.clone();
            //if (obj.syncsubparam != null) syncsubparam = (Concat)obj.syncsubparam.clone();
            if (obj.isActive)
            {
                _Mode = obj.mode;
                
                if (obj.inLegacy)       //Legacy
                {
                    if (obj._hostCachedProgram != null) _hostCachedProgram = (Prog)obj._hostCachedProgram.clone();
                    _hostNextPP = obj._hostNextPP;
                    _hostPP = obj._hostPP;
                }
                else                    //Virtualization Enabled
                {
                    //guest states
                    _guestNextPP = obj._guestNextPP;
                    _guestPP = obj._guestPP;
                    _guestPTP = obj._guestPTP;
                    _guestHVEntry = obj.HVEntry;
                    if (obj._guestCachedProgram != null) _guestCachedProgram = (Prog)obj._guestCachedProgram.clone();
                    //host states
                    if (obj.inHost)
                    {
                        if (obj._hostCachedProgram != null) _hostCachedProgram = (Prog)obj._hostCachedProgram.clone();
                        _hostNextPP = obj._hostNextPP;
                        _hostPP = obj._hostPP;
                        //vmexit states
                        if (obj.vmexitreason != VMExitReason.Default)
                        {
                            vmexitreason = obj.vmexitreason;
                            if (obj.vmexitreason == VMExitReason.Hypercall) hypercallid = obj.hypercallid;
                        }
                    }
                }

            }
        }

        public override string ToString()
        {
            //TODO
            return ExpressionID;
        }

        public override ExpressionValue GetClone()
        {
            return new Processor(this);
        }

        public override string ExpressionID
        {
            get
            {
                StringBuilder sb = new StringBuilder(); ;
                
                sb.Append("ps=" + _pagesize);
                //sync
                sb.Append(", syncState=" + syncstate);
                if (syncparam != null) sb.Append(", syncParam=" + syncparam.ExpressionID);
                //if (syncsubparam != null) sb.Append(", syncSubParam=" + syncsubparam.ExpressionID);
                if (isActive)
                {
                    sb.Append(mode);
                    if (inLegacy)       //Legacy
                    {
                        sb.AppendLine();
                        sb.Append("pnp=" + _hostNextPP);
                        sb.Append(", pp=" + _hostPP);
                        if (_hostCachedProgram != null) sb.Append(", prog=" + _hostCachedProgram.ExpressionID);
                        //sb.AppendLine();
                    }
                    else                    //Virtualization Enabled
                    {
                        //guest states
                        sb.Append(" G: ");
                        sb.Append("pnp=" + _guestNextPP);
                        sb.Append(", pp=" + _guestPP);
                        sb.Append(", ptp=" + _guestPTP);
                        sb.Append(", hv=" + _guestHVEntry);
                        if (_guestCachedProgram != null) sb.Append(", prog=" + _guestCachedProgram.ExpressionID);
                        
                        //host states
                        if (inHost)
                        {
                            sb.AppendLine();
                            sb.Append(" H: ");
                            sb.Append("pnp=" + _hostNextPP);
                            sb.Append(", pp=" + _hostPP);
                            if (_hostCachedProgram != null) sb.Append(", prog=" + _hostCachedProgram.ExpressionID);
                           
                            if (vmexitreason != VMExitReason.Default)
                            {
                                sb.AppendLine();
                                sb.Append(" VMExit: reason=" + vmexitreason);
                                if (vmexitreason == VMExitReason.Hypercall) sb.Append(", hypercallid =" + hypercallid);
                                //sb.AppendLine();
                            }
                        }
                    }
                }
                else
                {
                    //sb.Append("Inactive");
                    //sb.Append(" ps="+pagesize);
                    //sb.AppendLine();
                }
                return sb.ToString();
            }
        }

        public GDBase readReg(int id)
        {

            if (id == (int)Reg.guestPTPointer)
            {
                return new Value(guestPTP);
            }
            else if (id == (int)Reg.ProgramPointer)
            {
                return new Value(PP);
            }
            else if (id == (int)Reg.HypervisorEntry)
            {
                return new Value(HVEntry);
            }
            else
                throw new SysError();
        }

        public void writeReg(int id, GDBase data)
        {
            if (id == (int)Reg.guestPTPointer)
            {
                if (data is Value)
                    _guestPTP = ((Value)data).v;
                else
                    throw new SysError();
            }
            else
                throw new SysError();
        }

        public void reset()
        {
            _State = State.Inactive;
        }


        public void activateLegacy(int entry = 0)
        {
            _State = State.Active;
            _Mode = Mode.Legacy;
            _hostNextPP = entry;
            _hostCachedProgram = null;
            _hostPP = -1;
        }

        public void activateGuest(int hventry, int gptp, int gentry)
        {
            _State = State.Active;
            _Mode = Mode.Guest;
            _guestHVEntry = hventry;
            _guestPTP = gptp;
            _guestNextPP = gentry;
            _guestPP = -1;
        }

        public void halt()
        {
            _State = State.Inactive;
            syncPending(SyncException.SyncReason.Halt);
        }

        public void synchronized()
        {
            syncstate = SyncException.SyncReason.Nothing;
            syncparam = null;
        }

        public bool isSynchronized { get { return syncstate == SyncException.SyncReason.Nothing; } }

        public void syncPending(SyncException.SyncReason state, Concat param=null)
        {
            syncstate = state;
            syncparam = param;
        }

        public void syncPending(SyncException se)
        {
            syncPending(se.state, se.param);
        }
        
        public void VMExit(VMExitReason reason, int hcid=0)
        {
            _Mode = Mode.Host;
            _hostNextPP = _guestHVEntry;
            //vmexit
            vmexitreason = reason;
            if (reason == VMExitReason.Hypercall)  hypercallid = hcid;
            syncPending(SyncException.SyncReason.VMExit);
        }

        public void VMEntry()
        {
            _Mode = Mode.Guest;
            syncPending(SyncException.SyncReason.VMEntry);
        }

        /*
        public void VMExit_VON(int entry)
        {
            syncPending(SyncException.syncVMExit_VON());
            VMExit(entry);
        }

        public void VMExit_VOFF(int entry)
        {
            syncPending(SyncException.syncVMExit_VOFF());
            VMExit(entry);
        }
         * 
        public void VMExit_VMRESUME(int entry)
        {
            syncPending(SyncException.syncVMExit_VMRESUME());
            VMExit(entry);
        }

        public void VMExit_NEWVM(int entry)
        {
            syncPending(SyncException.syncVMExit_NEWVM());
            VMExit(entry);
        }

        public void VMExit_HYPERCALL(int entry, int callid, Concat userdata)
        {
            syncPending(SyncException.syncVMExit_HYPERCALL(callid, userdata));
            hypercallid = callid;
            VMExit(entry);
        }

        public void VMExit_LL(int entry)
        {
            syncPending(SyncException.syncVMExit_LL());
            VMExit(entry);
        }

        public void VMExit_Crash(int entry, Concat crashparam)
        {
            syncPending(SyncException.syncVMExit_Crash(crashparam));
            VMExit(entry);
        }
        */

    }
}
